/***************************************************************************
*
* bindiff:
*
*	Binary diff. Compare binary files and check if they are different.
*
* Author: 
*	Pallav Nawani.
*
***************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#define READBLK 4096
#define OPTSIZE 1024

typedef enum {
  FOLLOWARG,	/* An argument must follow this option */
  MAYFOLARG,	/* An argument may or may not follow this option */
  STRINGARG,	/* Not an option, must be removed by the user*/
  NOSUCHARG,	/* No such option exists */
  HIDDENARG,	/* Hidden option */
  LONGERARG,	/* Longer name option - One which starts from -- */
  NORMALARG,	/* Normal option */
 } Argtype;

/* To use the option types, do the following:
   FOLLOWARG	place a : after the option.
   MAYFOLARG	place a ; after the option.
   STRINGARG	Nothing needs to be done, the function returns this to tell
   		you that something which is not an option was encountered.
		
   NOSUCHARG	Nothing needs to be done, the function returns this to tell
   		you that an option which is not recognized was encountered.
   HIDDENARG	place a | after the option.
   LONGERARG	Put the full string (with -- in fullopt). This is a value
   		returned by the func to tell you that a long argument was
		encountered.

*/

/* Argument Structure returned by parse args */
typedef struct {
  int type;
  int code;
  } Opts;


char *fullopt[] = {
	 "--help",
	 "--verbose",
	 "--firstdiff",
	 "--offset:",
	 "--blockwise",
	 "--hexadecimal",
	 "*"
	 };

char *bopt[] = {
         "h",
	 "v",
	 "f",
	 "o:",
	 "b",
	 "x",
	 "*"
	 };

char *helps[] = {
  "Prints this help",
  "Lists each and every byte that differs between the binary files",
  "Exit after the first difference (default).",
  "Start Checking at this byte offset in both files. May be negat-
   \t\tive. If so, the offset is taken from the end of the file.",
  "Show the differences block wise. The way this works is this.
   \t\tSuppose two files differ at byte 16, and the differences
   \t\tcontinue until byte 23, and after that the two files are
   \t\tidentical until some other point. This mode will show a
   \t\tdifferences start line at byte 16 and differences end line
   \t\tat byte 24, and the program will continue on.",
  "Show the differences in hexadecimal format"
  };

struct _flags {
  int help;		/* Show the help */
  int verbose;		/* Show all the differences */
  int firstdiff;	/* Show first difference and exit - default */
  int offset;		/* Start diffing from this offset */
  int blkwise;		/* Show difference blockwise. Find the place where
  			   differences start, indicate the place, and continue
			   again find next set of differences and continue
			   and so on*/
  int hex;
  int done;		/* Check if we are done diffing */
  };

typedef struct _flags flags;


void usage(int i);

void error_exit(char *s) {
  fprintf(stderr, "bindiff: %s\n", s);
  fflush(stderr);
  exit(0);
  }

/************************************************************************
*
* print_args()
*	Print the arguments to the program to stdout. The arguments
*	are supposed to be in standard argc, argv format.
*
*************************************************************************/
void print_args(int argc, char **argv)
{
  int i;
  printf ("Program Arguments\n");
  for(i=0; i< argc; i++) 
    printf ("\t%s\n",argv[i]);
  fflush(stdout);  
  return;
}


/************************************************************************
*
* arglen()
*	Returns the real length of the option. That is, minus any
*	special characters such as :; etc.
*
*************************************************************************/
int arglen (char *arg) {
  int i;
  for(i=0; *arg!=';'&&*arg!=':'&&*arg!='|'&&*arg!='\0'; i++)
    arg++;

  return i;
  
  
}


/************************************************************************
*
* argtype()
*	Given an option string, returns its type, based on the special
*	characters present/not present in the option.
*************************************************************************/
int argtype ( char *arg) {
  int i;
  i = arglen(arg);
  
  if(arg[0]=='\0') return NOSUCHARG;
  else if(arg[i]==';') return MAYFOLARG;
  else if(arg[i]==':') return FOLLOWARG;
  else if(arg[i]=='|') return HIDDENARG;
  else if (arg[0]=='-'&&arg[1]=='-') return LONGERARG;
  else return NORMALARG;
  
  }

/************************************************************************
*
* help_args()
*	Takes as input the fulloptions table, the options table,
*	and the help table, and prints help messages on stdout describing
*	what each option does. It makes certain assumptions about
*	the order of the options and help strings. First, the order of
*	help strings must be the same as the order of the option strings,
*	each help string corresponding to the same index of he bopt
*	strings. In the same manner, fullopt option strings should
*	also correspond to the bopt string table element at the same
*	index.
*************************************************************************/
void help_args(char **fullopt, char **bopt, char **helps)
{
  char ptr[OPTSIZE], ptr2[OPTSIZE];
  int i, allsize;
  
  /* Calculate the number of options in bopt & store it in allsize */
  for(allsize=0; bopt[allsize][0]!='*';allsize++);

  for(i=0; bopt[i][0]!='*'; i++) {

        if(bopt[i][0]=='\0') continue;
    switch( argtype(bopt[i])) {
       case HIDDENARG:
       		break;
       
       case MAYFOLARG:
                strncpy(ptr, bopt[i], arglen(bopt[i]));
                printf("\t-%s\t%s\n", ptr, helps[i]);
		printf("\t\tThis arg MAY be followed by data\n\n");
		break;
		
       case FOLLOWARG:
                strncpy(ptr, bopt[i], arglen(bopt[i]));
                printf("\t-%s\t%s\n", ptr, helps[i]);
		printf("\t\tThis arg MUST be followed by required data\n\n");
		break;
       default:
                strncpy(ptr, bopt[i], arglen(bopt[i]));
                printf("\t-%s\t%s\n\n", ptr, helps[i]);
		break;
	}
     }

  for(i=0; fullopt[i][0]!='*'; i++) {
    if(fullopt[i][0]=='\0'||bopt[i]=='\0') continue;
    switch( argtype(fullopt[i])) {
       case HIDDENARG:
       		break;
       
       default: 
                strncpy(ptr, bopt[i], arglen(bopt[i]));
                strncpy(ptr2, fullopt[i], arglen(fullopt[i]));
		ptr2[arglen(fullopt[i])]=0;
                printf("\t%s\n\t\tLonger version of %s\n",ptr2, ptr);
				fflush(stdout);

		break;
	}
     }
  putchar('\n');

}


/***********************************************************************
*
* remove_top()
*    remove the top most elements of an argc, argv type argument list.
*    Value of argc will be decreased by 1 and top most element of the
*    argv array ( which is at the index 0 ) will be removed.
*
***********************************************************************/
void remove_top(int *argc, char **argv)
{
  int c,i;
  char *cc;
  char **narg;
  int sum=0;
  
  c=*argc;
  if(c==0) return;

  *argc = c-1;
  if(c==1) {
    free(argv[0]);
    return;
    }

  for(i=1; i<c; i++) 
    sum+=(strlen(argv[i])+1);
  cc = (char *)malloc(sum);
  
  narg = (char **)malloc(sizeof(char *)*(*argc));
  for(i=0,sum=0; i<c-1; i++) {
    narg[i]=cc;
    strcpy(cc, argv[i+1]);
    cc+=(strlen(argv[i+1])+1);
    }
  
  /* This assumes that all the arguments are in fact in a linear array */
  free(argv[0]);
  for(i=0; i<c-1; i++) argv[i] = narg[i];
  free(narg);
  return;
  
  }




/***********************************************************************
*
* parse_args()
*    Parse the argument list given in argc, argv format. This function
*    takes the following parameters. The argument count argc, the
*    argument array argv, the fullopt array, which contains the full
*    version of options, the bopt array, which contains the brief
*    version of the options.
*
*    After parsing the arguments, this function returns a token, which
*    can then be used to determine which option was passed. Each call
*    to this function parses and removes one option, and returns the
*    corresponding token, with one exception. If this function
*    encounters an argument which is not an option, it returns without
*    removing the argument. The user of this function then must decide
*    what that argument means and process it accordingly. After that
*    the user must remove the argument form the args list, and this
*    can be done by the utility function remove_top().
*
*    And remember, DO NOT free the pointer returned by parse_args!
*
***********************************************************************/
Opts *parse_args(int *argc, char **argv,
                            char **fullopt,
			    char **bopt)
{ 
  char *p1;
  int i;
  int done=0;
  static int inopt=0;
  static int first=0;
  static int k=0;
  int fsize=0;
  
  static Opts opto;
  opto.type = NOSUCHARG;
  opto.code = -1;
  
  /* Calculate the number of options in bopt & store it in fsize */
  for(fsize=0; bopt[fsize][0]!='*';fsize++);


  if(first==0) {
    first=1;
    }
  
  
  p1=*argv;
  do {
     switch(*p1) {
       
       case '-':  
                  if(inopt==1) {
		     if(k==1) {
		       k = inopt = 0;
		       opto.type = NOSUCHARG;
		       opto.code = -1;
		       return &opto;
		       }
		     p1--;
		     for(i=0; ;i++) {
		       if((strcmp(fullopt[i], "*")==0)) {
		         k = inopt = 0;
		         opto.type=NOSUCHARG;
			 opto.code = -1;
			 break;
			 }
		       if((strncmp(fullopt[i], p1,arglen(p1))==0)) {
			 remove_top(argc, argv);
			 opto.type=argtype(fullopt[i]);
			 opto.code = i+fsize;
		         break;
			 }
		       }
		     inopt=0;  
		     return &opto;
		     }
		   else {
		     inopt=1;
		     }  
		   break;
		   
       default:   if(inopt==1) {
       		     for(i=0; ;i++) {
		       if(strcmp(bopt[i],"*")==0) {
		          opto.code=-1;
			  opto.type=NOSUCHARG;
			  inopt = k =0 ;
			  break;
			  }
		       if(strncmp(bopt[i],p1,arglen(bopt[i]))==0) {
		          if(*(p1+1)=='\0') {
			    remove_top(argc, argv);
			    inopt = k = 0;
			    }
			  else {
			    strcpy(*argv,p1+arglen(bopt[i]));
			    k=1;
			    }
			  opto.code=i;
			  opto.type=argtype(bopt[i]);
			  break;
			  }
		       }
		     return &opto;
		     }
		   
		   else {
		     opto.code=-1;
		     opto.type=STRINGARG;
		     return &opto;
		     }
		   break;
      
      }

   p1++;

   }while(!done);

  /* Just to pacify gcc */
  opto.type = NOSUCHARG;
  opto.code = -1;
  return &opto;    

}



void usage(int i) {
 printf("\nBindiff is a program to find differences between two binary ");
 printf("files. It\nindicates the byte(s) at which difference(s) were found.");
 printf("\n\nUsage: [options] filename1 filename2\n\n");
 help_args(fullopt, bopt, helps);
 if(i==1) exit(-1);
}



int main(int argc, char **argv) {
  
  char data[1024];
  unsigned char blk1[READBLK], blk2[READBLK];
  char *fnames[2];
  int i =0, m, k=0, n1, n2;
  int offset;
  int set =0;
  flags flg;
  FILE *inpf1, *inpf2;
  Opts *optr;
  
  
    /* Initialize data */
  k = m = n1 = n2 = 0;
  inpf1 = inpf2 = NULL;
  memset(((void *)&flg), 0, sizeof(flags));
  
    /* check arguments */
  if(argc<2) {
    printf("Bindiff: Error! Incorrect Number of Arguments!");
    usage(1);
    }

  if(argc==2) {
     usage(1);
     }
  
  /* Remove the topmost argument pair, it is just the program name*/
  remove_top(&argc, argv);
  /* Parse the arguments */
  do{

    optr=parse_args(&argc, argv, fullopt, bopt);
/*
    print_args(argc, argv);
    printf ("optr->code %d, optr->type %d, argc%d\n",optr->code, optr->type, argc);
*/   
    switch(optr->type) {

       case STRINGARG :  
                        if(i>1) {
			   printf("Error! too many filenames!\n");
			   usage(1);
			   }
                        fnames[i++] = strdup(argv[0]);
			remove_top(&argc, argv);
			break;
			
      case FOLLOWARG :	flg.offset = 1;
      			errno = 0;
			if(argc>0) offset = atoi(argv[0]);
			else errno =1;
			if(errno) usage(1);
			remove_top(&argc, argv);
			break;

      case MAYFOLARG :	/* fall through */
      case HIDDENARG :	/* fall through */
      case LONGERARG :	/* fall through */
      case NORMALARG :
	                
			switch(optr->code) {
    

      			case 0:
                           usage(1);
			   break;
      			case 1:
                           flg.verbose =1;
			   break;
      			case 2:
                          flg.firstdiff = 1;
			  break;
      			case 3:
                          flg.offset = 1;
			  break;
      			case 4:
                          flg.blkwise = 1;
			  break;

      			case 5:
			  flg.hex = 1;
                          break;
			  
      			case 6:
                           usage(1);
			   break;
      			case 7:
                           flg.verbose =1;
			   break;
      			case 8:
                          flg.firstdiff = 1;
			  break;
      			case 9:
                          flg.offset = 1;
			  break;
      			case 10:
                          flg.blkwise = 1;
			  break;

      			case 11:
			  flg.hex = 1;
                          break;
			
      		        default:
			  printf("Error! Unrecognized Option!\n");
      			  usage(1);
			  break;

      
                        }

			break;
			
      case NOSUCHARG:   /* fall through */
      default:		
                        printf("Error! Unrecognized Option!\n");
      			usage(1);
			break;

     
      }
      
      
    } while(k!= NOSUCHARG && argc !=0);
  
  /* set default flag */
  if(!flg.verbose&&!flg.blkwise&&!flg.blkwise) flg.firstdiff=1;
  
  if((inpf1=fopen(fnames[0], "r"))==NULL) {
    sprintf(data, "Unable to open file %s :%s.", fnames[0], strerror(errno));
    error_exit(data);
    }
  
  if((inpf2=fopen(fnames[1], "r"))==NULL) {
    sprintf(data, "Unable to open file %s :%s.", fnames[1], strerror(errno));
    error_exit(data);
    }
  
  if(flg.offset!=0) {
    
    /* where to seek from: file start or file end*/

    i=(offset<0?SEEK_END:SEEK_SET);
    fseek(inpf1, 0, SEEK_END);
    fseek(inpf2, 0, SEEK_END);
    n1=ftell(inpf1);
    n2=ftell(inpf2);
    fseek(inpf2, 0, SEEK_SET);
    fseek(inpf1, 0, SEEK_SET);
    
    if(n1<abs(offset)||n2<abs(offset)) 
      error_exit("Offsets bigger than the file size!\n");

    if((fseek(inpf1, offset, i))==-1||
       (fseek(inpf2, offset, i))==-1)
     {
      perror("Bindiff: Error in using offset!");
      fclose(inpf1); fclose(inpf2);
      usage(1);
      }

    }
    
  k=0;
  if(flg.offset) k+= offset;
    
  while(!flg.done) {
    n1=fread(blk1, 1, READBLK, inpf1);
    n2=fread(blk2, 1, READBLK, inpf2);
    m =(n1>n2?n2:n1);
    
    for(i=0; i<m; k++, i++) {

      if(blk1[i]^blk2[i]) {
        
	if(flg.verbose||flg.firstdiff||(flg.blkwise&&!set)) {
	   if(!flg.hex)
	     printf("[%d] :\t%u\t%u\n", k, blk1[i],blk2[i]);
	   else
	     printf("[%x] :\t%x\t%x\n", k, blk1[i],blk2[i]);
	   if(flg.blkwise&&!set) {
	     set=1;
	     }
	   }

        if((!flg.verbose&&!flg.blkwise) || flg.firstdiff) {
	  fflush(stdout);
	  flg.done=1;
	  break;
	  }	
	}

      else {
	 if(flg.blkwise&&set) {
	   if(!flg.hex)
	     printf("<%d> :\t%u\t%u\n\n", k, blk1[i],blk2[i]);
	   else
	     printf("<%x> :\t%x\t%x\n\n", k, blk1[i],blk2[i]);
	   set=0;
	   }
	 } 
	
      }

    if(m<READBLK) flg.done=1;
    
    
    }
  
  fclose(inpf1);
  fclose(inpf2);
  
  return 0;
  }
